#include <windows.h>
#include <stdlib.h>
#include <stdio.h>

#include "DXWCPU.h"

#define MHZ_MILLIS	(500)	// 0.005 accurate

DXWCPU::DXWCPU(void)
{
	m_CPUGotInfo		= false;

	m_CPU_Features		= 0;
	m_CPU_MHZ			= 0;
	m_CPU_PerformanceFreq = 1.0f;
	m_CPU_SecondsPerClock = 1.0f;

	m_MaxCPUIDVal		= 0;
	m_ProcType			= 0;
	m_ProcFamily		= 0;
	m_ProcModel			= 0;
	m_ProcStepping		= 0;

	uint32	m_L2Cache	= 0;
	
	GetInfo();
}

//--------------------------------------------------------------------------------------------------------------------------------
// CPUID routins
//--------------------------------------------------------------------------------------------------------------------------------

// Execute CPUID function EAX=funcNum and get EAX result
uint32 DXWCPU::GetCPUID_EAX(uint32 funcNum)
{
	uint32 retval;

	__try
	{
		_asm
		{
			mov eax, funcNum
			CPUID
			mov retval, eax
		}
	}
	__except(EXCEPTION_EXECUTE_HANDLER)
	{
		retval = 0;
	}
	return retval;
}

// Execute CPUID function EAX=funcNum and get EDX result
uint32 DXWCPU::GetCPUID_EDX(uint32 funcNum)
{
	uint32 retval;

	__try
	{
		_asm
		{
			mov eax, funcNum
			CPUID
			mov retval, edx
		}
	}
	__except(EXCEPTION_EXECUTE_HANDLER)
	{
		retval = 0;
	}
	return retval;
}

// Execute CPUID EAX=funcNum and get EAX result and
// the string EBX:EDX:ECX
uint32 DXWCPU::GetCPUIDString(uint32 funcNum, char *String)
{
	uint32 retval;

	__try
	{
		_asm
		{
			mov	eax,funcNum
			CPUID
			mov	retval,eax
			mov	eax,String
			mov	dword ptr[eax],ebx
			mov	dword ptr[eax+4],edx
			mov	dword ptr[eax+8],ecx
		}
	}__except(EXCEPTION_EXECUTE_HANDLER)
	{
		retval	=0;
	}

	return	retval;

}

// AMD Specific:
// In the AMD processor extended EAX functions 8000_0002h, 8000_0003h, 
//  and 8000_0004h return an ASCII string containing the name of the processor.
void DXWCPU::GetCPUIDStringAMD(uint32 funcNum, char *String)
{
	uint32	retval;
	
	__try
	{
		_asm
		{
			mov eax,funcNum
			CPUID
			mov	retval,eax
			mov	eax,String
			mov	dword ptr[eax+4],ebx
			mov	dword ptr[eax+8],ecx
			mov	ebx,retval
			mov	dword ptr[eax+12],edx
			mov	dword ptr[eax],ebx
		}
	}
	__except(EXCEPTION_EXECUTE_HANDLER)
	{
		retval	=0;
	}
}

int64 DXWCPU::GetRDTSC(void)
{
	int64 clock;

	_asm
	{
		push	ebx
		push	ecx
		xor		eax,eax
		CPUID
		RDTSC
		mov		dword ptr clock,eax
		mov		dword ptr clock+4,edx
		xor		eax,eax
		CPUID
		pop		ecx
		pop		ebx
	}
	return	clock;
}

uint32 DXWCPU::GetRDTSCOverHead(void)
{
	uint32	Elap, MinOverhead	=0xffffff;
	int64	start;
	int		x;

	for(x=0;x < 50;x++)
	{
		start		=GetRDTSC();
		Elap		=(uint32)(GetRDTSC() - start);
		MinOverhead =min(MinOverhead, Elap);
	}
	return	MinOverhead;
}

//--------------------------------------------------------------------------------------------------------------------------------
// MHZ measuring routins
//--------------------------------------------------------------------------------------------------------------------------------
static int64 StartClock;
static uint32 StartTime;

uint32 RoundMHZ(uint32 MHZ)
{
	const uint32 NF[9] = {0, 20, 33, 50, 60, 66, 80, 90, 100};
  
	uint32 Freq, RF;
	uint8 i;
	uint8 Hi, Lo;
	
	RF = 0;
	Freq = MHZ % 100;
	for (i = 0; i<8; i++)
	{
		if (Freq < NF[i])
		{
			Hi = i;
			Lo = i - 1;
			if ((NF[Hi] - Freq) > (Freq - NF[Lo])) 
				RF = NF[Lo] - Freq;
			else
				RF = NF[Hi] - Freq;
			
			break;
		}
	}

  return MHZ + RF;
}

void DXWCPU::StartGetMHZ(void)
{
	StartClock	= GetRDTSC();
	StartTime	= timeGetTime();	
}

uint32 DXWCPU::GetMHZ(void)
{
	int64	ElapsedClock;
	uint32	ElapsedTime, MHZ;

	do
	{
		ElapsedTime = timeGetTime() - StartTime;
	}
	while (ElapsedTime < MHZ_MILLIS);

	ElapsedClock = GetRDTSC() - StartClock;
	MHZ = (uint32)(ElapsedClock/(1000*ElapsedTime));

	// Check for errors
//	uint32 Z50,Z83;

/*	Z50 = ((MHZ + 25)/50)*50; // round to the nearest fifty
	Z83 = (int)((MHZ + 41)*12/1000)*1000/12; // round to the nearest 83.333 == 1000/12*/
		
	MHZ=RoundMHZ(MHZ);

	return MHZ;
}

bool DXWCPU::GetPerformanceFreq(void)
{
	LARGE_INTEGER Freq;

	if (!QueryPerformanceFrequency(&Freq))
	{
		return false;
	}

	m_CPU_PerformanceFreq = (float)Freq.LowPart;

	return true;
}

//--------------------------------------------------------------------------------------------------------------------------------
bool DXWCPU::GetInfo(void)
{
	uint32	TypeFlags = 0;

	if (m_CPUGotInfo) 
		return true;
	m_CPUGotInfo = true;

	StartGetMHZ();

	// Get the CPU vendor string using the standard EAX=0
	// function. the CPUID instruction will return a 12chars 
	// vendor string
	ZeroMemory(m_ProcVendorString,sizeof(m_ProcVendorString));
	m_MaxCPUIDVal	= GetCPUIDString(0, m_ProcVendorString);
	
	if ((strncmp(m_ProcVendorString, VENDOR_INTEL, 12))==0)
	{
		// CPU vendor is Intel

		// Get features flag
		TypeFlags = GetCPUID_EDX(0x1);
		m_CPU_Features = TypeFlags;

		TypeFlags		= GetCPUID_EAX(1);
		m_ProcType		= (TypeFlags>>12)	& 0x3;
		m_ProcFamily	= (TypeFlags>>8)	& 0xf;
		m_ProcModel		= (TypeFlags>>4)	& 0xf;
		m_ProcStepping	= (TypeFlags)		& 0x7;

		sprintf(m_ProcName,GetIntelProcName(m_ProcType, m_ProcFamily, m_ProcModel, m_ProcStepping));
	}
	else if ((strncmp(m_ProcVendorString, VENDOR_AMD, 12))==0)
	{
		TypeFlags = GetCPUID_EAX(0x80000000);

		if (TypeFlags) // Extended functions supported
		{
			TypeFlags = GetCPUID_EDX(0x80000001);
			m_CPU_Features = TypeFlags;

			GetCPUIDStringAMD(0x80000002, m_ProcName);
			GetCPUIDStringAMD(0x80000003, m_ProcName+16);
			GetCPUIDStringAMD(0x80000004, m_ProcName+32);
		}
		else
		{
			TypeFlags = GetCPUID_EDX(0x1);
			m_CPU_Features = TypeFlags;
		}
		TypeFlags		= GetCPUID_EAX(1);
		m_ProcType		=(TypeFlags>>12)	& 0x3;
		m_ProcFamily	=(TypeFlags>>8)		& 0xf;
		m_ProcModel		=(TypeFlags>>4)		& 0xf;
		m_ProcStepping	=(TypeFlags)		& 0x7;
	}

	GetPerformanceFreq();

	m_CPU_MHZ = GetMHZ();
	m_CPU_SecondsPerClock = 1.0f / ( 1000000.0f * m_CPU_MHZ );

	return true;
}

char* DXWCPU::GetIntelProcName(uint32 Type, uint32 Family, uint32 Model, uint32 Stepping)
{
	char *CPUName = NULL;

	switch (Family)
	{
		case 4: if (Model == 4) CPUName = INTEL_486SL; 
				if (Model == 7) CPUName = INTEL_DX2WB; 
				if (Model == 8) CPUName = INTEL_DX4; 
				if ((Model == 8)&&(Type==1)) CPUName = INTEL_DX4O;
				break;
		
		case 5:	if (Model == 1) CPUName = INTEL_P;
				if (Model == 2) CPUName = INTEL_P;
				if ((Model == 1)&&(Type==1)) CPUName = INTEL_PO;
				if ((Model == 2)&&(Type==1)) CPUName = INTEL_PO;
				if ((Model == 3)&&(Type==1)) CPUName = INTEL_PO;
				if (Model == 3) CPUName = INTEL_P; 
				if (Model == 4) CPUName = INTEL_P; 
				if ((Model == 4)&&(Type==1)) CPUName = INTEL_PO; 
				break;
				
		case 6: if (Model == 1) CPUName = INTEL_PPRO;
				if (Model == 3) CPUName = INTEL_PII; 
				if ((Model == 3)&&(Type=1)) CPUName = INTEL_PIIO;
				if (Model == 5) CPUName = INTEL_PII;
				if (Model == 6) CPUName = INTEL_CELERON;
				if (Model == 7) CPUName = INTEL_PIII; 
				if (Model == 8) CPUName = INTEL_PIII; 
				if (Model == 10) CPUName = INTEL_XEON;
				break;

		default: CPUName = PROCESSOR_UNKNOWN; break;
				
	}
	return CPUName;
}

void DXWCPU::GetIntelCacheDescription(uint8 D, char* Desc)
{
	switch (D)
	{
		case 0x01 : sprintf(Desc,"Instruction TLB, 4Kb pages, 4-way set associative, 32 entries"); break;
		case 0x02 : sprintf(Desc,"Instruction TLB, 4Mb pages, fully associative, 2 entries"); break;
		case 0x03 : sprintf(Desc,"Data TLB, 4Kb pages, 4-way set associative, 64 entries"); break;
		case 0x04 : sprintf(Desc,"Data TLB, 4Mb pages, 4-way set associative, 8 entries"); break;
		case 0x06 : sprintf(Desc,"8KB instruction cache, 4-way set associative, 32 byte line size"); break;
		case 0x08 : sprintf(Desc,"16KB instruction cache, 4-way set associative, 32 byte line size"); break;
		case 0x0A : sprintf(Desc,"8KB data cache 2-way set associative, 32 byte line size"); break;
		case 0x0C : sprintf(Desc,"16KB data cache, 4-way set associative, 32 byte line size"); break;
		case 0x40 : sprintf(Desc,"No L2 cache"); break;
		case 0x41 : sprintf(Desc,"Unified cache, 32 byte cache line, 4-way set associative, 128Kb"); break;
		case 0x42 : sprintf(Desc,"Unified cache, 32 byte cache line, 4-way set associative, 256Kb"); break;
		case 0x43 : sprintf(Desc,"Unified cache, 32 byte cache line, 4-way set associative, 512Kb"); break;
		case 0x44 : sprintf(Desc,"Unified cache, 32 byte cache line, 4-way set associative, 1Mb"); break;
		case 0x45 : sprintf(Desc,"Unified cache, 32 byte cache line, 4-way set associative, 2Mb"); break;
	}
}



